/***********************************************************************************************************************
 *
 * blueBill Stats
 * Copyright (C) 2011-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it/mobile/
 * SCM: http://java.net/hg/bluebill-server~stats-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.stats.domain.impl;

import javax.annotation.Nonnull;
import javax.annotation.PostConstruct;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.zip.GZIPInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URLDecoder;
import org.joda.time.format.DateTimeFormatter;
import org.joda.time.format.DateTimeFormatterBuilder;
import it.tidalwave.bluebill.stats.domain.Ping;
import it.tidalwave.bluebill.stats.domain.PingDao;
import it.tidalwave.bluebill.stats.domain.PingFinder;
import lombok.Cleanup;
import lombok.Getter;
import lombok.Setter;
import lombok.extern.slf4j.Slf4j;

/***********************************************************************************************************************
 *
 * An in-memory implementation of {@link PingDao} which can be populated by reading an http server log file.
 * 
 * @stereotype DAO
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
@Slf4j
public class HttpServerLogPingDao implements PingDao
  {
    private static final DateTimeFormatter FORMATTER = new DateTimeFormatterBuilder().appendDayOfMonth(2)
                                                                                     .appendLiteral("/")
                                                                                     .appendMonthOfYearShortText()
                                                                                     .appendLiteral("/")
                                                                                     .appendYear(4, 4)
                                                                                     .appendLiteral(":")
                                                                                     .appendHourOfDay(2)
                                                                                     .appendLiteral(":")
                                                                                     .appendMinuteOfHour(2)
                                                                                     .appendLiteral(":")
                                                                                     .appendSecondOfMinute(2)
                                                                                     .appendLiteral(" ")
                                                                                     .appendTimeZoneOffset("", false, 2, 2)
                                                                                     .toFormatter();
    /** The in-memory data. */
    private final List<Ping> pings = new ArrayList<Ping>();
    
    /** The list of file paths to read data from. */
    @Getter @Setter
    private List<String> fileNames = new ArrayList<String>();
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     * 
     ******************************************************************************************************************/
    @Override
    public void insert (final @Nonnull Ping ping)
      {
        log.info("insert()", ping);
        pings.add(ping);            
      }
    
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     * 
     ******************************************************************************************************************/
    @Override @Nonnull
    public PingFinder findPings() 
      {
        return new HttpServerLogPingFinderSupport() 
          {
            @Override @Nonnull
            protected List<? extends Ping> computeResults()
              {
                return new ArrayList<Ping>(pings);
              }
          };
      }    
    
    /*******************************************************************************************************************
     *
     * Initializes the DAO by reading data.
     * 
     ******************************************************************************************************************/
    @PostConstruct
    /* package */ void initialize()
      {
        log.info("initialize()");
        
        for (final String fileName : fileNames)
          {
            final File file = new File(fileName);
            
            if (file.exists())
              {
                try 
                  {
                    load(file);
                  }
                catch (IOException e) 
                  {
                    log.error("Cannot load from file: " + file, e);
                  }
              }
          }
      }
    
    /*******************************************************************************************************************
     *
     * Loads data from the given file.
     * 
     * @param  file   the file to read data from
     * 
     ******************************************************************************************************************/
    private void load (final @Nonnull File file) 
      throws IOException
      {
        log.info("load({})", file);
        InputStream is = new FileInputStream(file);
        
        if (file.getName().endsWith(".gz"))
          {
            is = new GZIPInputStream(is);
          }
        
        final @Cleanup BufferedReader br = new BufferedReader(new InputStreamReader(is));
        
//        final Pattern pattern = Pattern.compile("^(.*) - - \\[(.*)\\].*$");
            
        for (;;)
          {
            final String s = br.readLine();
            
            if (s == null)
              {
                break;    
              }
            
//            log.trace(">>>> {}", s);
           
            if (s.contains("ping?"))
              {
                try
                  {
                    // TODO: regexp works in unit test, but not in the webapp ?!?!?
//                    final Matcher matcher = pattern.matcher(s);
//                    assert matcher.matches() : "no match";
//                    final String ipAddress = matcher.group(1);
//                    final String timeStampAsString = matcher.group(2);
                    final int i1 = s.indexOf(" ");
                    final String ipAddress = s.substring(0, i1);
                    final int i2 = s.indexOf("[");
                    final int i3 = s.indexOf("]");
                    final String timeStampAsString = s.substring(i2 + 1, i3);

                    final String queryString = s.replaceAll("^.*GET /mobile/ping\\?", "").replaceAll(" .*$", "");
                    log.trace(">>>> {}", queryString);      

                    final Map<String, String> map = new HashMap<String, String>();

                    for (final String nameValuePair : queryString.split("&"))
                      {
                        final String[] splitNameValuePair = nameValuePair.split("=");
                        final String propertyName = URLDecoder.decode(splitNameValuePair[0], "UTF-8");
                        final String propertyValue = URLDecoder.decode(splitNameValuePair[1], "UTF-8");
//                        log.trace(">>>>>>>> {} = {}", propertyName, propertyValue);
                        map.put(propertyName, propertyValue);
                      }

                    pings.add(new Ping(FORMATTER.parseDateTime(timeStampAsString),
                                       ipAddress,
                                       map.get("installId"),
                                       map.get("blueBillVersion").replace(".null", "").replace("null", ""), 
                                       map.get("osVersion"), 
                                       map.get("javaVersion"), 
                                       map.get("userLanguage"), 
                                       map.get("userRegion"), 
                                       map.get("board"), 
                                       map.get("brand"), 
                                       map.get("device"), 
                                       map.get("model"), 
                                       map.get("product"), 
                                       map.get("versionIncremental"), 
                                       map.get("versionRelease"), 
                                       map.get("versionSDK")));
                  }
                catch (IllegalStateException e)
                  {
                    throw new RuntimeException("Cannot parse " + s, e);  
                  }
              }
          }
      }
  }
